#ifndef STRUCTS_H
#define STRUCTS_H

#include "api-driver/stricts-enable.h"
#include "proxy.h"
#include <atomic>
#include <fstream>
#include <shared_mutex>
#include <string>
#include <boost/log/trivial.hpp>
#include <boost/property_tree/ptree_fwd.hpp>



namespace api_driver::obj {
    /**
     * Обертка для объектов, доступных для обновления
     * NOTE: перед вызовом функций, требующих `TSID`, необходимо захватить мютекс API.
     */
    class CpUpdatebleObject {
    public:
        int64_t lastUpdate = 0;
        int64_t updatePeriodMs = -1;

        /**
         * Функция для обновления (загрузки) объекта из CP API.
         */
        virtual void updateCallback(proxy::CpProxy& cp) = 0;

        bool checkNeedUpdate(int64_t now) const;

        int64_t getNextUpdate(int64_t now) const;

        virtual ~CpUpdatebleObject();
    };


#if API_OBJECT_STATISTICS_ENABLE
    class StatisticsLogger: public CpUpdatebleObject {
    public:
        StatisticsLogger();

        static constexpr const char* LOG_FILENAME = "/tmp/weblog-statistics.csv";

        int64_t timeStart;

        bool logEn = false;
        std::atomic<int> logPeriodMs = 1000;
        std::atomic<int> maxAgeMs = 10000;

        /**
         *
         * @return {"en": bool, "logPeriodMs": int, "maxAgeMs": int}
         */
        std::string getSettings();
        void setSettings(boost::property_tree::ptree &pt);

        void updateCallback(proxy::CpProxy& cp) override;
        /**
         * Записать значение в "базу данных". Метку при этом вставлять не нужно, она будет вставлена автоматически.
         * @param item
         */
        void putItem(const debug_metrics& item);

        // void collectExpiredItems();

        // void resetLogs() {
        //     std::lock_guard _lock(mutex);
        //     logs.clear();
        // }

        ~StatisticsLogger() override;
    private:
        // std::pmr::deque<LogItem> logs;
        std::fstream logFile{};
        std::shared_mutex mutex;
    };
#endif

#if defined(MODEM_IS_SCPC)
    static constexpr const char* DEFAULT_SERVER_NAME = "RSCM-101";
#elif defined(MODEM_IS_TDMA)
    static constexpr const char* DEFAULT_SERVER_NAME = "TDMA Abonent";
#elif defined(MODEM_IS_SHPC)
    static constexpr const char* DEFAULT_SERVER_NAME = "SHPS Terminal";
#else
#error "Selected modem type not supported!"
#endif

#if API_OBJECT_NETWORK_SETTINGS_ENABLE
    class TerminalNetworkSettings: public CpUpdatebleObject {
    public:
        std::string managementIp, managementGateway, dataIp, serverName;
        bool isL2 = true;
        unsigned int dataMtu = 1500;

        TerminalNetworkSettings();
        TerminalNetworkSettings(const TerminalNetworkSettings& src);
        TerminalNetworkSettings& operator= (const TerminalNetworkSettings& src);

        void loadDefaults();

        void updateCallback(proxy::CpProxy& cp) override;
        void updateFromPt(boost::property_tree::ptree &pt);
        void store(proxy::CpProxy& cp);
        std::string asJson();

        ~TerminalNetworkSettings() override;
    };
#endif

#if API_OBJECT_QOS_SETTINGS_ENABLE
    class TerminalQosSettings: public CpUpdatebleObject {
    public:
        static constexpr const char* DEFAULT_QOS_CLASSES = R"({"rt1":[],"rt2":[],"rt3":[],"cd":[]})";
        std::string qosSettingsJson;
        bool qosEnabled = false;

        TerminalQosSettings();
        TerminalQosSettings(const TerminalQosSettings& src);
        TerminalQosSettings& operator= (const TerminalQosSettings& src);

        void loadDefaults();

        void updateCallback(proxy::CpProxy& cp) override;
        void updateFromPt(boost::property_tree::ptree &pt);
        void store(proxy::CpProxy& cp);
        std::string asJson();

        ~TerminalQosSettings() override;
    };
#endif

    class TerminalFirmwareVersion {
    public:
        std::string version, modemId, modemSn, macMang, macData;

        TerminalFirmwareVersion();
        TerminalFirmwareVersion(const TerminalFirmwareVersion& src);
        ~TerminalFirmwareVersion();

        // logCpApiError("api_driver::TerminalApiDaemon::connectToApi()->CP_GetNetwork(version)", CP_GetNetwork(sid, "version", &f.version));
        // logCpApiError("api_driver::TerminalApiDaemon::connectToApi()->CP_GetNetwork(chip_id)", CP_GetNetwork(sid, "chip_id", &f.modemId));
        // rtrim(f.modemId);
        // logCpApiError("api_driver::TerminalApiDaemon::connectToApi()->CP_GetNetwork(serial)", CP_GetNetwork(sid, "serial", &f.modemSn));
        // logCpApiError("api_driver::TerminalApiDaemon::connectToApi()->CP_GetNetwork(mac_eth0)", CP_GetNetwork(sid, "mac_eth0", &f.macMang));
        // logCpApiError("api_driver::TerminalApiDaemon::connectToApi()->CP_GetNetwork(mac_eth1)", CP_GetNetwork(sid, "mac_eth1", &f.macData));
        void load(proxy::CpProxy& cp);
        std::string asJson();

        TerminalFirmwareVersion& operator= (const TerminalFirmwareVersion& src);
    };

    /**
     * Обертка состояния терминала, тут состояние девайса, модулятора/демодулятора, состояние автоматического обновления.
     */
    class TerminalState: public CpUpdatebleObject {
    public:
        std::string fInitState{};
        bool fIsTest = false; // daemon->isTest()
        bool fTxState = false;
#ifdef MODEM_IS_SCPC
        bool fIsCinC = false;
#endif

        bool fStatRxState;
        bool fStatRxSymSyncLock;
        bool fStatRxFreqSearchLock;
        bool fStatRxAfcLock;
        bool fStatRxPktSync;

        float fStatRxSnr;
        float fStatRxRssi;
        uint16_t fStatRxModcod;
        bool fStatRxFrameSizeNormal;
        bool fStatRxIsPilots;

        double fStatRxSymError;
        double fStatRxFreqErr;
        double fStatRxFreqErrAcc;
        double fStatRxInputSignalLevel;
        double fStatRxPllError;
        double fStatRxSpeedOnRxKbit;
        double fStatRxSpeedOnIifKbit;
        uint32_t fStatRxPacketsOk;
        uint32_t fStatRxPacketsBad;
        uint32_t fStatRxPacketsDummy;

        bool fStatTxState;
        uint16_t fStatTxModcod;
        double fStatTxSpeedOnTxKbit;
        double fStatTxSpeedOnIifKbit;
#ifdef MODEM_IS_SCPC
        float fStatTxSnr;
        bool fStatTxFrameSizeNormal;
        bool fStatTxIsPilots;
        double fStatCincOcc;
        bool fStatCincCorrelator;
        uint32_t fStatCincCorrelatorFails;
        int32_t fStatCincFreqErr;
        int32_t fStatCincFreqErrAcc;
        float fStatCincChannelDelay;
#endif
#ifdef MODEM_IS_TDMA
        fStatTxCenterFreq;
        fStatTxSymSpeed;
#endif

        /**
         * Обновление основной части статистики, то есть RX/TX и sysinfo
         */
        void updateCallback(proxy::CpProxy& cp) override;

        void updateFromPt(boost::property_tree::ptree &pt);
        void store(TSID sid, CP_Result& lastCpError);
        std::string asJson();

        ~TerminalState() override;
    };

    class TerminalDeviceState: public CpUpdatebleObject {
    public:
        time_t fOsUptime{};
        double fOsLoad1{};
        double fOsLoad5{};
        double fOsLoad15{};
        unsigned int fOsTotalram{};
        unsigned int fOsFreeram{};
        unsigned int fOsProcs{};

        double fTempAdrv{};
        double fTempZynq{};
        double fTempFpga{};
#ifdef MODEM_IS_TDMA
        DOWNLOAD_STATUS fUpgradeStatus{};
        unsigned int fUpgradePercent{};
        std::string fUpgradeImage;
#endif

        TerminalDeviceState();
        TerminalDeviceState(const TerminalDeviceState& src);
        TerminalDeviceState& operator= (const TerminalDeviceState& src);

        void updateCallback(proxy::CpProxy& cp) override;
        std::string asJson() const;

        ~TerminalDeviceState() override;
    };

    class TerminalRxTxSettings: public CpUpdatebleObject {
    public:

        // TODO описать все параметры для всех терминалов

        void updateCallback(proxy::CpProxy& cp) override;
        void updateFromPt(boost::property_tree::ptree &pt);
        void store(proxy::CpProxy& cp);
        std::string asJson();

        ~TerminalRxTxSettings() override;
    };
}

#endif //STRUCTS_H
